/*****************************************************************************
;
;   (C) ADMtek Incorporated.  All Rights Reserved.
;
;------------------------------------------------------------------------------
;
;    Project : ADM5120
;    Creator : daniell@admtek.com.tw
;    File    : arch/mips/adm5120/irq.c
;	 Date    : 2003.3.4
;    Abstract: 
;
;Modification History:
;

;*****************************************************************************/

#include <linux/init.h>
#include <linux/kernel_stat.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/random.h>
#include <linux/pm.h>

#include <asm/irq.h>
#include <asm/mipsregs.h>
#include <asm/gdb-stub.h>
#include <asm/irq_cpu.h>

#include <asm/adm5120/adm5120.h>


extern void breakpoint(void);
extern int setup_irq(unsigned int irq, struct irqaction *irqaction);
extern irq_desc_t irq_desc[];
extern asmlinkage void mipsIRQ(void);

static void enable_adm5120_irq(unsigned int irq);
static void disable_adm5120_irq(unsigned int irq);


/*
 * irq enable
 */
static __inline void adm5120_irq_enable(int adm_irql)
{
	unsigned long s;

	save_and_cli(s);
	ADM5120_INTC_REG(IRQ_ENABLE_REG) = (1 << adm_irql);
	restore_flags(s);
}


/*
 * irq disable
 */
static __inline void adm5120_irq_disable(int adm_irql)
{
	unsigned long s;

	save_and_cli(s);
	ADM5120_INTC_REG(IRQ_DISABLE_REG) = (1 << adm_irql);
	restore_flags(s);
}


/*
 * startup irq
 */
unsigned int startup_adm5120_irq(unsigned int irq)
{
	int adm_irq;

	adm_irq = irq - ADM5120_IRQ_OFFSET;
	if ((adm_irq < 0) || (adm_irq > INT_LVL_MAX)) 
		return -1;

	adm5120_irq_enable(adm_irq);

	return 0;
}


/*
 * shutdown irq
 */
void shutdown_adm5120_irq(unsigned int irq)
{
	int adm_irq;

	adm_irq = irq - ADM5120_IRQ_OFFSET;
	if ((adm_irq < 0) || (adm_irq > INT_LVL_MAX)) 
		return;

	adm5120_irq_disable(adm_irq);
}
	

/* 
 * enable 5120 irq 
 */
void enable_adm5120_irq(unsigned int irq)
{
	int adm_irq;

	adm_irq = irq - ADM5120_IRQ_OFFSET;
	if ((adm_irq < 0) || (adm_irq > INT_LVL_MAX)) 
		return;

	adm5120_irq_enable(adm_irq);
}


/*
 * disable 5120 irq 
 */
void disable_adm5120_irq(unsigned int irq)
{
	int adm_irq;

	adm_irq = irq - ADM5120_IRQ_OFFSET;
	if ((adm_irq < 0) || (adm_irq > INT_LVL_MAX)) 
		return;

	adm5120_irq_disable(adm_irq);
}


/*
 * 5120 irq ack
 */
static void ack_adm5120_irq(unsigned int irq_nr)
{
	int adm_irq;

	adm_irq = irq_nr - ADM5120_IRQ_OFFSET;
	if ((adm_irq < 0) || (adm_irq > INT_LVL_MAX)) 
		return;

	ADM5120_INTC_REG(IRQ_DISABLE_REG) = (1 << adm_irq);
}


/*
 * 5120 irq end
 */
static void end_adm5120_irq(unsigned int irq_nr)
{
	int adm_irq;

	adm_irq = irq_nr - ADM5120_IRQ_OFFSET;
	if ((adm_irq < 0) || (adm_irq > INT_LVL_MAX)) 
		return;

	ADM5120_INTC_REG(IRQ_ENABLE_REG) = (1 << adm_irq);
}


/*
 * 5120 irq set affinity 
 */
void set_affinity_adm5120_irq(unsigned int irq, unsigned long mask)
{
	return;
}

	
/*
 * system irq type
 */
static hw_irq_controller adm5120_irq_type = {
	.typename = "ADM5120 INTC",
	.startup = startup_adm5120_irq,
	.shutdown = shutdown_adm5120_irq,
	.enable = enable_adm5120_irq,
	.disable = disable_adm5120_irq,
	.ack = ack_adm5120_irq,
	.end = end_adm5120_irq,
	.set_affinity = set_affinity_adm5120_irq
};


/*
 * irq init
 */
void __init init_adm5120_irqs(void)
{
	int i;

	for (i = ADM5120_IRQ_OFFSET; i <= ADM5120_IRQ_MAX; i++) {
		irq_desc[i].status = IRQ_DISABLED;
		irq_desc[i].action = NULL;
		irq_desc[i].depth = 1;
		irq_desc[i].handler = &adm5120_irq_type;
	}
}


/*
 * system init
 */
void __init init_IRQ(void)
{
	set_except_vector(0, mipsIRQ);
	
	init_generic_irq();

	mips_cpu_irq_init(0);
	
	init_adm5120_irqs();

	set_c0_status(ALLINTS);

#ifdef CONFIG_REMOTE_DEBUG
 	printk("Setting debug traps - please connect the remote debugger.\n");

	set_debug_traps();

	breakpoint();
#endif
}

static spinlock_t mips_irq_lock = SPIN_LOCK_UNLOCKED;

/* 
 * System irq dispatch
 */
asmlinkage void adm5120_hw0_irqdispatch(struct pt_regs *regs)
{
	unsigned long flags;
	unsigned long intsrc;
	int i;

	spin_lock_irqsave(&mips_irq_lock, flags);

	intsrc = ADM5120_INTC_REG(IRQ_STATUS_REG) & IRQ_MASK;

	for (i = 0; intsrc; intsrc >>= 1, i++) {
		if (intsrc & 0x1)
			do_IRQ(ADM5120_IRQ_OFFSET + i, regs);
	}

	spin_unlock_irqrestore(&mips_irq_lock, flags);
}

