/*****************************************************************************
;
;   (C) ADMtek Incorporated.  All Rights Reserved.
;
;------------------------------------------------------------------------------
;
;    Project : ADM5120
;    Creator : 
;    File    : 
;    Abstract: 
;
;Modification History:
; 
;
;*****************************************************************************/
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/types.h>
#include <asm/byteorder.h>
#include <asm/pci_channel.h>
#include <asm/adm5120/adm5120.h>

volatile u32* pci_config_address_reg = (volatile u32*)KSEG1ADDR(0x115ffff0);
volatile u32* pci_config_data_reg = (volatile u32*)KSEG1ADDR(0x115ffff8);

#define PCI_ENABLE 0x80000000
                             
#define cfgaddr(dev, where) (((dev->bus->number & 0xff) << 0x10) |  \
                             ((dev->devfn & 0xff) << 0x08) |        \
                             (where & 0xfc)) | PCI_ENABLE


/*
 * We can't address 8 and 16 bit words directly.  Instead we have to
 * read/write a 32bit word and mask/modify the data we actually want.
 */
static int adm5120_read_config_byte (struct pci_dev *dev,
                                   int where, unsigned char *val)
{	
	*pci_config_address_reg = cfgaddr(dev, where);
    
#ifdef CONFIG_CPU_LITTLE_ENDIAN
	*val = (le32_to_cpu(*pci_config_data_reg) >> ((where&3)<<3)) & 0xff;
#else
	*val = ((*pci_config_data_reg) >> ((where&3)<<3)) & 0xff;
#endif
    
//	printk("pci_read_byte 0x%x == 0x%x\n", where, *val);
	return PCIBIOS_SUCCESSFUL;
}

static int adm5120_read_config_word (struct pci_dev *dev,
                                   int where, unsigned short *val)
{
	if (where & 1)
		return PCIBIOS_BAD_REGISTER_NUMBER;
		
	*pci_config_address_reg = cfgaddr(dev, where);
	
#ifdef CONFIG_CPU_LITTLE_ENDIAN
	*val = (le32_to_cpu(*pci_config_data_reg) >> ((where&3)<<3)) & 0xffff;
#else
	*val = ((*pci_config_data_reg) >> ((where&3)<<3)) & 0xffff;
#endif

//	printk("pci_read_word 0x%x == 0x%x\n", where, *val);
	return PCIBIOS_SUCCESSFUL;
}

int adm5120_read_config_dword (struct pci_dev *dev,
                                    int where, unsigned int *val)
{
	if (where & 3)
		return PCIBIOS_BAD_REGISTER_NUMBER;
	*pci_config_address_reg = cfgaddr(dev, where);
	
#ifdef CONFIG_CPU_LITTLE_ENDIAN
	*val = le32_to_cpu(*pci_config_data_reg);
#else
	*val = (*pci_config_data_reg);
#endif

//	printk("pci_read_dword 0x%x == 0x%x\n", where, *val);
	return PCIBIOS_SUCCESSFUL;
}

static int adm5120_write_config_byte (struct pci_dev *dev,
                                    int where, unsigned char val)
{
	*pci_config_address_reg = cfgaddr(dev, where);
	*(volatile u8 *)(((int)pci_config_data_reg) + (where & 3)) = val;
//	printk("pci_write_byte 0x%x = 0x%x\n", where, val);
	return PCIBIOS_SUCCESSFUL;
}

static int adm5120_write_config_word (struct pci_dev *dev,
                                    int where, unsigned short val)
{
	if (where & 1)
		return PCIBIOS_BAD_REGISTER_NUMBER;
	*pci_config_address_reg = cfgaddr(dev, where);
	*(volatile u16 *)(((int)pci_config_data_reg) + (where & 2)) = (val);
//	printk("pci_write_word 0x%x = 0x%x\n", where, val);
	return PCIBIOS_SUCCESSFUL;
}

static int adm5120_write_config_dword (struct pci_dev *dev,
                                     int where, unsigned int val)
{
	if (where & 3)
		return PCIBIOS_BAD_REGISTER_NUMBER;

	*pci_config_address_reg = cfgaddr(dev, where);
	*pci_config_data_reg = (val);
//	printk("pci_write_dword 0x%x = 0x%x\n", where, val);
	return PCIBIOS_SUCCESSFUL;
}



struct pci_ops adm5120_pci_ops = {
	adm5120_read_config_byte,
	adm5120_read_config_word,
	adm5120_read_config_dword,
	adm5120_write_config_byte,
	adm5120_write_config_word,
	adm5120_write_config_dword
};


struct resource pciioport_resource = {
	"pci IO space", 
	0x11500000,  
	0x115ffff0-1,
	IORESOURCE_IO
};

struct resource pciiomem_resource = {
	"pci memory space", 
	0x11400000,
	0x11500000-1,
	IORESOURCE_MEM
};


struct pci_channel mips_pci_channels[] = {
	{ &adm5120_pci_ops, &pciioport_resource, &pciiomem_resource, 0, 0xff},
	{(struct pci_ops *)NULL, (struct resource *)NULL,
	 (struct resource *)NULL, (int)0, (int)0}
};



unsigned __init int pcibios_assign_all_busses(void)
{
        return 1;
}

