/*
 * Carsten Langgaard, carstenl@mips.com
 * Copyright (C) 1999,2000 MIPS Technologies, Inc.  All rights reserved.
 *
 * ########################################################################
 *
 *  This program is free software; you can distribute it and/or modify it
 *  under the terms of the GNU General Public License (Version 2) as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * ########################################################################
 *
 * Setting up the clock on the MIPS boards.
 *
 */

#include <linux/types.h>
#include <linux/config.h>
#include <linux/init.h>
#include <linux/kernel_stat.h>
#include <linux/sched.h>
#include <linux/spinlock.h>

#include <asm/mipsregs.h>
#include <asm/ptrace.h>
#include <asm/hardirq.h>
#include <asm/div64.h>
#include <asm/cpu.h>
#include <asm/adm5120/adm5120.h>
#include <linux/interrupt.h>
#include <linux/timex.h>
#include <asm/time.h>



void __init adm5120_time_init(void)
{
	unsigned long clock;

	clock = (ADM5120_SW_REG(CODE_REG) & CODE_CLK_MASK) >> CODE_CLK_SHIFT;

	switch (clock)
	{
	case CPU_CLK_175MHZ:
		mips_hpt_frequency = CPU_SPEED_175M;
		printk("CPU clock: 175MHz\n");
		break;

	case CPU_CLK_200MHZ:
		mips_hpt_frequency = CPU_SPEED_200M;
		printk("CPU clock: 200MHz\n");
		break;

	case CPU_CLK_225MHZ:
		mips_hpt_frequency = CPU_SPEED_225M;
		printk("CPU clock: 225MHz\n");
		break;

	case CPU_CLK_250MHZ:
		mips_hpt_frequency = CPU_SPEED_250M;
		printk("CPU clock: 250MHz\n");
		break;
	}
}



void __init adm5120_timer_setup(struct irqaction *irq)
{
    /* to generate the first timer interrupt */
	write_c0_compare(read_c0_count() + mips_hpt_frequency);	

	setup_irq(MIPS_CPU_TIMER_IRQ, irq);
}

