#include <linux/init.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/bootmem.h>
#include <asm/bootinfo.h>
#include <asm/addrspace.h>
#include <asm/adm5120/prom.h>

void __init prom_init_cmdline(int argc, char **argv);
void __init prom_init_env(char **env);

char arcs_cmdline[CL_SIZE];
char env_buf[PROM_ENVSIZE];
char *prom_envp[MAXPROMENV];


int __init
prom_init(int argc, char **argv, char **envp, int *prom_vec)
{
	unsigned char *memsize_str;
	unsigned long memsize;

	mips_machgroup = MACH_GROUP_ADM5120;
	mips_machtype = MACH_COREGA_LBSTD2PU;

	prom_init_cmdline(argc, argv);
	prom_init_env(envp);
	memsize_str = prom_getenv("memsize");
	if (memsize_str == NULL) 
		memsize_str = prom_getenv("ramsize");
	if (!memsize_str) {
		memsize = 0x04000000;
	} else {
		memsize = simple_strtol(memsize_str, NULL, 0);
	}
	add_memory_region(0, memsize, BOOT_MEM_RAM);

	return 0;
}



char *
prom_getenv(char *envname)
{
	char **env = prom_envp;
	int len;

	len = strlen(envname);

	while (*env) {
		if (strncmp(envname, *env, len) == 0)
			return *env + len + 1;
		env++;
	}

	return NULL;
}



char *
prom_getcmdline(void)
{
	return &(arcs_cmdline[0]);
}



/*
 * prom: init command line
 */
void __init
prom_init_cmdline(int argc, char **argv)
{
	char *cp, *ep;
	int i, n;

	cp = &arcs_cmdline[0];
	ep = &arcs_cmdline[CL_SIZE - 1];

	for (i = 1; i < argc; i++) { /* ignore argv[0] */
		n = ep - cp;
		if (n == 0)
			break;
		strncpy(cp, argv[i], n);
		cp[n] = 0;
		cp = strchr(cp, 0);
		*cp++ = ' ';
	}

	if (cp != &(arcs_cmdline[0])) /* get rid of trailing space */
		--cp;
	
	*cp = '\0';
}


void __init
prom_init_env(char **env)
{
	char *cp, *ep;
	int i, n;

	cp = &env_buf[0];
	ep = &env_buf[PROM_ENVSIZE - 1];

	if (env) {
		for (i = 0; env[i] && i < MAXPROMENV - 1; i++) {
			n = ep - cp;
			if (n == 0)
				break;
			strncpy(cp, env[i], n);
			prom_envp[i] = cp;
			cp[n] = 0;
			cp = strchr(cp, 0);
			cp++;
		}
	}

	prom_envp[i] = 0;
}



void __init
prom_free_prom_memory(void)
{
}



int
prom_set_ethernet_address(int port, unsigned char *mac)
{
	return -1;
}



int
prom_get_ethernet_address(int port, unsigned char *mac)
{
	char *p;
	char envname[] = "ethaddrXX";
	int i;

	if (port > 6)
		return -1;

	snprintf(envname, sizeof(envname), "ethaddr%d", port);
	p = prom_getenv(envname);
	if (p == NULL) {
		if (port == 0)
			p = prom_getenv("ethaddr");
		else {
			snprintf(envname, sizeof(envname), "eth%daddr", port);
			p = prom_getenv(envname);
		}
		if (p == NULL)
			return -1;
	}

	for (i = 0; i < 6; i++) {
		if (*p == 0)
			break;
		mac[i] = (unsigned char)simple_strtoul(p, &p, 0x10);
		if (*p == ':' && *p == '.' && *p != 0)
			break;
		if (*p != 0)
			p++;
	}
	if (i == 6)
		return 0;

	return -1;
}



int
prom_get_vlan_groups(unsigned char *vlan)
{
	char *p;
	int i;

	p = prom_getenv("ethports");
	/*
	 * Corega LBSTD2PU VLAN ethports default setting Changed..
	 * 2007.11.05  T.Yoshida
	 * VLAN0 : 0x41 , VLAN1 : 0x42
	 */
#if 0
	if (p == NULL) {
		vlan[0] = 0x40 | 0x01 | 0x02 | 0x04 | 0x08;	/* lan side */
		vlan[1] = 0x40 | 0x10;	/* wan side */
		vlan[2] = 0;
		vlan[3] = 0;
		vlan[4] = 0;
		vlan[5] = 0;
		return 2;
	}	
#else
	if (p == NULL) {
		vlan[0] = 0x40 | 0x01;			/* lan Side */
		vlan[1] = 0x40 | 0x02;			/* wan side */
		vlan[2] = 0;
		vlan[3] = 0;
		vlan[4] = 0;
		vlan[5] = 0;
		return 2;
	}	
#endif


	i = 0;
	while (*p) {
		unsigned char ports;
		if (*p == '0' && *(p + 1) == 'x') {
			p += 2;
			if (*p == 0)
				break;
		}
		ports = (unsigned char)simple_strtoul(p, &p, 0x10);
		if (ports != 0)
			vlan[i++] = ports; 
		if (*p == 0)
			break;
		if (*p != ',' && *p != ':')
			break;
		p++;
	}

	return i;
}

