/*****************************************************************************
;
;   (C) ADMtek Incorporated.  All Rights Reserved.
;
;------------------------------------------------------------------------------
;
;    Project : ADM5120
;    Creator : daniell@admtek.com.tw
;    File    : net/adm5120sw/if5120.c
;	 Date    : 2003.5.14
;    Abstract: ADM5120 Switch driver
;
;Modification History:
;
;*****************************************************************************/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>

#include <linux/sched.h>
#include <linux/kernel.h>       /* printk() */
#include <linux/slab.h>         /* kmalloc() */
#include <linux/errno.h>        /* error codes */
#include <linux/types.h>        /* size_t */
#include <linux/interrupt.h>    /* mark_bh */

#include <linux/in.h>
#include <linux/netdevice.h>    /* struct device, and other headers */
#include <linux/etherdevice.h>  /* eth_type_trans */
#include <linux/ip.h>           /* struct iphdr */
#include <linux/tcp.h>          /* struct tcphdr */
#include <linux/skbuff.h>
#include <linux/in6.h>
#include <linux/proc_fs.h>
#include <linux/ctype.h>

#include <asm/checksum.h>
#include <asm/adm5120/adm5120.h>

#include "if5120.h"

static struct proc_dir_entry *adm5120sw_dir;
static struct proc_dir_entry *adm5120sw_port_dir[NUM_IF5120_PORTS];

#if 0
#define ADM5120SW_DEBUG   1
#endif

extern int swdrv_init(int, unsigned char *);
extern void swdrv_ProcessInt(int irq, void *dev_id, struct pt_regs *regs);
extern int swdrv_IfOpen(PSW_PRIV_T priv, char *mac);
extern int swdrv_IfClose(PSW_PRIV_T priv);
extern int swdrv_HwTx(struct sk_buff *skb, PSW_PRIV_T priv);
extern void EnablePromisc(int unit, unsigned long rxmode);
extern int SetupVLAN(int unit, unsigned long portmask);
extern int ProgramVlanMac(int vlan, unsigned char *Mac, int clear);
extern void EnableVlanGroup(int unit, unsigned long vlanGrp);
extern void DisableVlanGroup(int unit, unsigned long vlanGrp);
extern void UpdateLinkStatus(void);

#ifdef CONFIG_PROC_FS
static int adm5120sw_procfs_init(void);
#endif

int adm5120sw_init(struct net_device *dev);

/* This is a load-time options */
static int timeout = 5;
static int irqEn = 0;
static int unit = 0;

extern PSW_CONTEXT_T sw_context;

/*
 * The devices
 */
struct net_device adm5120sw_devs[ADM5120SW_IFNUM_MAX] =
{
    { name:ADM5120SW_IF_PREFIX "0", init:adm5120sw_init },
    { name:ADM5120SW_IF_PREFIX "1", init:adm5120sw_init },
    { name:ADM5120SW_IF_PREFIX "2", init:adm5120sw_init },
    { name:ADM5120SW_IF_PREFIX "3", init:adm5120sw_init },
    { name:ADM5120SW_IF_PREFIX "4", init:adm5120sw_init },
    { name:ADM5120SW_IF_PREFIX "5", init:adm5120sw_init }
};

static int nvlangroups = 0;
static unsigned char vlangroup[MAX_VLAN_GROUP];


static int vlan_to_port(int vlan_num, unsigned *port_bit)
{
	unsigned vlan = (sw_context->vlanGrp[vlan_num]) & 0x3f;
	unsigned bit;
	int port_num;

	bit = 1;
	for (port_num = 0; port_num < NUM_IF5120_PORTS; port_num++) {
		if ((bit & vlan) != 0)
			break;
		bit <<= 1;
	}

	if ((vlan & ~bit) != 0) {
		/* the vlan has many ports */
		return -1;
	}

	if (port_bit != NULL)
		*port_bit = bit;

	return port_num;
}



/*
 
 * Open and close
 
 */

int adm5120sw_open(struct net_device *dev)

{
	PSW_PRIV_T priv = (PSW_PRIV_T)dev->priv;
	unsigned long flags;
	unsigned port_bit, tmp;
	int port_num;
	int unit = priv->unit;
	int retval;

#ifdef ADM5120SW_DEBUG
	printk("Open if: %s\n", dev->name);
#endif

	/* enable switch irq routine */
	if (!irqEn) {
		retval = request_irq(SWITCH_IRQ,
			swdrv_ProcessInt, SA_SHIRQ, dev->name, dev);
		if (retval)
			return -ENODEV;

		irqEn = 1;
	}

	spin_lock_irqsave(&sw_context->lock, flags);

	SetupVLAN(unit, sw_context->vlanGrp[unit]);

	/* program vlan mac */
	ProgramVlanMac(unit, dev->dev_addr, 0);

	/* Enable vlan Group */
	EnableVlanGroup(unit, sw_context->vlanGrp[unit]);

 	priv->vlanId = 1 << unit;

	if (sw_context->actIfCnt == 0)
	{
		/* Enable interrupt */
		ADM5120_SW_REG(SW_Int_mask_REG) = ~sw_context->intMask; 
		adm_sync();
	}
	sw_context->actIfCnt++;

	dev->irq = SWITCH_IRQ;
	netif_start_queue(dev);
	dev->flags &= ~IFF_RUNNING;

	UpdateLinkStatus();

	port_num = vlan_to_port(unit, &port_bit);
	if (port_num != -1) {
		if (port_num == 5)
			tmp = ADM5120_SW_REG(Port_conf2_REG);
		else 
			tmp = ADM5120_SW_REG(PHY_cntl2_REG);
		if (tmp & 1)
			dev->flags |= IFF_AUTOMEDIA;
		else
			dev->flags &= ~IFF_AUTOMEDIA;
	} else
		dev->flags |= IFF_AUTOMEDIA;

	spin_unlock_irqrestore(&sw_context->lock, flags);

	return 0;
}



int adm5120sw_release(struct net_device *dev)

{
	PSW_PRIV_T priv = (PSW_PRIV_T)dev->priv;
	unsigned long flags;
	int unit = priv->unit;

	spin_lock_irqsave(&sw_context->lock, flags);

	netif_stop_queue(dev);      /* can't transmit any more */
	dev->flags &= ~IFF_RUNNING;

	/* disable vlan group */
	DisableVlanGroup(unit, sw_context->vlanGrp[unit]);

	if (--sw_context->actIfCnt <= 0)
	{
		ADM5120_SW_REG(SW_Int_mask_REG) = SWITCH_INT_MASK;
		adm_sync();
	}

	spin_unlock_irqrestore(&sw_context->lock, flags);

	return 0;
}



/*
 * Configuration changes (passed on by ifconfig)
 */

int adm5120sw_config(struct net_device *dev, struct ifmap *map)
{
	PSW_PRIV_T priv = (PSW_PRIV_T)dev->priv;
	unsigned port_bit, mask, tmp;
	unsigned flags;
	int port_num;
	int error = 0;;

	port_num = vlan_to_port(priv->unit, &port_bit);
	if (port_num == -1) {
		/* the vlan has many ports */
		return -EINVAL;
	}

	/* mask for phy_cntl2 */
	mask = (port_bit << 5) | (port_bit << 10) | (port_bit << 15);

	spin_lock_irqsave(&sw_context->lock, flags);

	switch (map->port) {
	case IF_PORT_UNKNOWN:
		if (port_num != 5) {
			error = -EINVAL;
		} else {
			ADM5120_SW_REG(Port_conf2_REG) =
				(ADM5120_SW_REG(Port_conf2_REG) & ~0x3f) | 0x3c;
		}
		break;
    
	case IF_PORT_10BASET:
		if (port_num != 5) {
			tmp = ADM5120_SW_REG(PHY_cntl2_REG) & ~mask;
			if (sw_context->port[port_num].duplex)
				tmp |= (port_bit << 10);
			if (sw_context->port[port_num].flow_control)
				tmp |= (port_bit << 15);
			ADM5120_SW_REG(PHY_cntl2_REG) = tmp;
		} else {
			ADM5120_SW_REG(Port_conf2_REG) =
				(ADM5120_SW_REG(Port_conf2_REG) & ~0x3f) | 0x38;
		}
		break;

	case IF_PORT_100BASETX:
		if (port_num != 5) {
			tmp = ADM5120_SW_REG(PHY_cntl2_REG) & ~mask;
			if (sw_context->port[port_num].duplex)
				tmp |= (port_bit << 10);
			if (sw_context->port[port_num].flow_control)
				tmp |= (port_bit << 15);
			tmp |= port_bit << 5;
			ADM5120_SW_REG(PHY_cntl2_REG) = tmp;
		} else {
			ADM5120_SW_REG(Port_conf2_REG) =
				(ADM5120_SW_REG(Port_conf2_REG) & ~0x3f) | 0x3a;
		}
		break;
	default:
		return -EINVAL;
	}

	UpdateLinkStatus();

	spin_unlock_irqrestore(&sw_context->lock, flags);
    
	return 0;
}


/*
 * Configuration changes (passed on by ifconfig)
 * Daniel+ 2003.8.4
 *
 * this is dev->set_multicast_list, but it is always called on SIOCSIFFLAGS.
 */
void adm5120sw_set_rx_mode(struct net_device *dev)
{
    PSW_PRIV_T priv = (PSW_PRIV_T)dev->priv;
    int port_num;
    unsigned port_bit, tmp, flags;

    /* Note: do not reorder, GCC is clever about common statements. */
    if (dev->flags & IFF_PROMISC)
    {
#ifdef ADM5120SW_DEBUG
        printk (KERN_NOTICE "%s: Promiscuous mode enabled.\n", dev->name);
#endif
        priv->iflags = IF_PROMISC_MODE;
    }
    else
    {
#ifdef ADM5120SW_DEBUG
        printk (KERN_NOTICE "%s: Promiscuous mode disabled.\n", dev->name);
#endif
        priv->iflags &= ~IF_PROMISC_MODE;
    }
    /* We can safely update without stopping the chip. */
    EnablePromisc(priv->unit, priv->iflags);

    port_num = vlan_to_port(priv->unit, &port_bit);
    if (port_num == -1)
	return;

    spin_lock_irqsave(&sw_context->lock, flags);

	if (port_num != 5) {
		tmp = ADM5120_SW_REG(PHY_cntl2_REG);
    		if (dev->flags & IFF_AUTOMEDIA) {
			if ((tmp & port_bit) == 0) 
				ADM5120_SW_REG(PHY_cntl2_REG) = tmp | port_bit;
		} else {
			if ((tmp & port_bit) != 0) 
				ADM5120_SW_REG(PHY_cntl2_REG) = tmp & ~port_bit;
		}
	} else {
		tmp = ADM5120_SW_REG(Port_conf2_REG);
    		if (dev->flags & IFF_AUTOMEDIA) {
			if ((tmp & 1) == 0)
				ADM5120_SW_REG(Port_conf2_REG) = tmp | 1;
		} else {
			if ((tmp & 1) != 0)
				ADM5120_SW_REG(Port_conf2_REG) = tmp & ~1;
		}
	}

    spin_unlock_irqrestore(&sw_context->lock, flags);
}



/*
 * Transmit a packet (called by the kernel)
 */
int adm5120sw_tx(struct sk_buff *skb, struct net_device *dev)
{
    
    PSW_PRIV_T priv = (PSW_PRIV_T)dev->priv;
    unsigned long eflags;

    if (skb == NULL) 
        return 0;

#ifdef GPIO_MEASURE
    GPIO_TOGGLE(2);
#endif

    spin_lock_irqsave(&priv->lock, eflags);

#ifdef GPIO_MEASURE
    GPIO_TOGGLE(2);
#endif


    dev->trans_start = jiffies; /* save the timestamp */

#ifdef GPIO_MEASURE
    GPIO_TOGGLE(2);
#endif

    /* transmit data to hardware descriptor */
    
    swdrv_HwTx(skb, priv);

#ifdef GPIO_MEASURE
    GPIO_TOGGLE(2);
#endif

#ifdef GPIO_MEASURE
    GPIO_TOGGLE(2);
#endif

    spin_unlock_irqrestore(&priv->lock, eflags);

#ifdef GPIO_MEASURE
    GPIO_TOGGLE(2);
#endif

    return 0;
}




/*
 * Deal with a transmit timeout.
 */
void adm5120sw_tx_timeout (struct net_device *dev)
{
    return;
}






/*
 * Ioctl commands 
 */
int adm5120sw_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
    return 0;
}



/*
 * Return statistics to the caller
 */
struct net_device_stats *adm5120sw_stats(struct net_device *dev)
{
    PSW_PRIV_T priv = (PSW_PRIV_T)dev->priv;
    
    return &priv->stats;
}



/*
 * This function is called to fix mac address
 */
int adm5120sw_set_mac(struct net_device *dev, void *p)
{
    PSW_PRIV_T priv = (PSW_PRIV_T)dev->priv;
    struct sockaddr *addr = (struct sockaddr *)p;
    unsigned long flags;
    spinlock_t *lock = &((PSW_PRIV_T)dev->priv)->lock;

    spin_lock_irqsave(lock, flags);

    memcpy(dev->dev_addr, addr->sa_data, dev->addr_len);

    prom_set_ethernet_address(priv->unit, dev->dev_addr);

    spin_unlock_irqrestore(lock, flags);

    return 0;
}


/*
 * This function is called to fill up an eth header, since arp is not
 * available on the interface
 */
int adm5120sw_rebuild_header(struct sk_buff *skb)
{
    return 0;
}



int adm5120sw_header(struct sk_buff *skb, struct net_device *dev,
                    unsigned short type, void *daddr, void *saddr,
                    unsigned int len)
{
    struct ethhdr *eth = (struct ethhdr *)skb_push(skb,ETH_HLEN);
    
    eth->h_proto = htons(type);
    
    memcpy(eth->h_source, saddr ? saddr : dev->dev_addr, dev->addr_len);
    memcpy(eth->h_dest,   daddr ? daddr : dev->dev_addr, dev->addr_len);
    
    return (dev->hard_header_len);
}





/*
 * The "change_mtu" method is usually not needed.
 * If you need it, it must be like this.
 */
int adm5120sw_change_mtu(struct net_device *dev, int new_mtu)
{
    unsigned long flags;
    
    spinlock_t *lock = &((PSW_PRIV_T)dev->priv)->lock;
    
    /* check ranges */
    
    if ((new_mtu < 68) || (new_mtu > 1500))
        
    return -EINVAL;

   /*
     
     * Do anything you need, and the accept the value
     
     */
    
    spin_lock_irqsave(lock, flags);
    
    dev->mtu = new_mtu;
    spin_unlock_irqrestore(lock, flags);
    
    return 0; /* success */
}



/*
 * The init function (sometimes called probe).
 * It is invoked by register_netdev()
 */
int adm5120sw_init(struct net_device *dev)
{
    PSW_PRIV_T priv;

    /* assign some of the fields */
    
    ether_setup(dev);
    
    dev->open            = adm5120sw_open;
    dev->stop            = adm5120sw_release;
    dev->set_config      = adm5120sw_config;
    dev->set_multicast_list = adm5120sw_set_rx_mode;
    dev->set_mac_address = adm5120sw_set_mac;
    dev->hard_start_xmit = adm5120sw_tx;
    dev->do_ioctl        = adm5120sw_ioctl;
    dev->get_stats       = adm5120sw_stats;
    dev->change_mtu      = adm5120sw_change_mtu;
    dev->rebuild_header  = adm5120sw_rebuild_header;
    dev->hard_header     = adm5120sw_header;
    dev->tx_timeout     = adm5120sw_tx_timeout;
    dev->watchdog_timeo = timeout;

    set_bit(__LINK_STATE_PRESENT, &dev->state);

    /*
     * Then, allocate the priv field. This encloses the statistics
     * and a few private fields.
     */
    dev->priv = kmalloc(sizeof(SW_PRIV_T), GFP_KERNEL);
    
    if (dev->priv == NULL)
        return (-ENOMEM);

    memset(dev->priv, 0, sizeof(SW_PRIV_T));

    priv = (PSW_PRIV_T)dev->priv;
    priv->unit = unit++;

    spin_lock_init(&priv->lock);

    return 0;
}



/*
 * module init
 */
static int __init adm5120switch_init(void)
{
	int result, i;

	printk(KERN_INFO "ADM5120 Switch Module: irq %d\n", SWITCH_IRQ);

	nvlangroups = prom_get_vlan_groups(vlangroup);
	if (nvlangroups <= 0) {
		if (nvlangroups == 0) {
			printk(KERN_INFO
			"ADM5120 Switch Module: no VLAN group configuration\n");
		}
		return -ENODEV;
	}

	printk(KERN_INFO "ADM5120 Switch Module: VLAN groups ");
	for (i = 0; i < nvlangroups; i++)
		printk("%d:%02x ", i, vlangroup[i]);
	printk("\n");

	/* Initialize the switch device */
	if (swdrv_init(nvlangroups, vlangroup) != 0) {
		return -ENODEV;
	}

	for (i = 0; i < nvlangroups;  i++) {
		if (prom_get_ethernet_address(i, adm5120sw_devs[i].dev_addr) ==
		    -1) {
			printk(KERN_WARNING
			    "%s: no mac address. use default address.\n",
			    adm5120sw_devs[i].name);
			memcpy(adm5120sw_devs[i].dev_addr,
			    ADM5120SW_DEFAULT_MAC0, 6);
			adm5120sw_devs[i].dev_addr[5] = i;
		}
		if ((result = register_netdev(adm5120sw_devs + i))) {
			printk(KERN_INFO
			    "adm5120sw: error %i registering device \"%s\"\n",
			    result, adm5120sw_devs[i].name);
			break;
		}
	}

	if (i != nvlangroups) {
		while (i) {
			kfree(adm5120sw_devs[i].priv);
			unregister_netdev(adm5120sw_devs + i);
			i--;
		}
		return -ENODEV;
	}

#ifdef CONFIG_PROC_FS
	adm5120sw_procfs_init();
#endif

	return 0;
}



#ifdef CONFIG_PROC_FS

#define _STATUS_ENABLE	0
#define _STATUS_VLAN	1
#define _STATUS_LINK	2
#define _STATUS_SPEED	3
#define _STATUS_DUPLEX	4
#define _STATUS_FC	5
#define _CONTROL_AN	0
#define _CONTROL_SPEED	1
#define _CONTROL_DUPLEX	2
#define _CONTROL_FC	3
#define _CONTROL_MDIX	4

#define _PROC_ID(port, entry)	((port << 16) | entry)
#define _PROC_ID_PORT(id)	((id  >> 16) & 0xffff)
#define _PROC_ID_ENTRY(id)	(id  & 0xffff)

static int proc_read_vlan(char *buf, char **start, off_t offset,
			  int len, int *eof, void *data);
static int proc_read_all_port_status(char *buf, char **start, off_t offset,
			  int len, int *eof, void *data);
static int proc_read_port_status(char *buf, char **start, off_t offset,
			  int len, int *eof, void *data);
static int proc_read_port_control(char *page, char **start, off_t offset,
			  int len, int *eof, void *data);
static int proc_write_port_control(struct file *file, const char *userbuf,
				 unsigned long count, void *data);
struct port_entry {
	char *name;
	mode_t mode;
	int entry;
	int (*read)(char *, char **, off_t, int, int *, void *);
	int (*write)(struct file *, const char *, unsigned long, void *);
} proc_port_entries[] = {
	{"status_enable", 0444, _STATUS_ENABLE, proc_read_port_status, NULL},
	{"status_link", 0444, _STATUS_LINK, proc_read_port_status, NULL},
	{"status_speed", 0444, _STATUS_SPEED, proc_read_port_status, NULL},
	{"status_duplex", 0444, _STATUS_DUPLEX, proc_read_port_status, NULL},
	{"status_flow_control", 0444, _STATUS_FC, proc_read_port_status, NULL},
	{"status_vlan", 0444, _STATUS_VLAN, proc_read_port_status, NULL},
	{"control_auto_negotiation", 0644, _CONTROL_AN,
		proc_read_port_control, proc_write_port_control},
	{"control_speed", 0644, _CONTROL_SPEED,
		proc_read_port_control, proc_write_port_control},
	{"control_duplex", 0644, _CONTROL_DUPLEX,
		proc_read_port_control, proc_write_port_control},
	{"control_flow_control", 0644, _CONTROL_FC,
		proc_read_port_control, proc_write_port_control},
	{"control_auto_mdix", 0644, _CONTROL_MDIX,
		proc_read_port_control, proc_write_port_control},
	{NULL}
};

static int
adm5120sw_procfs_init(void)
{
	int i, j;

	adm5120sw_dir = NULL;
	for (i = 0; i < NUM_IF5120_PORTS; i++)
		adm5120sw_port_dir[i] = NULL;

	adm5120sw_dir = proc_mkdir("adm5120sw", NULL);
	if (adm5120sw_dir == NULL) {
		printk(KERN_ERR "adm5120sw: proc_mkdir() failed\n");
		return 0;
	}
	create_proc_read_entry("vlan_ports", 0444, adm5120sw_dir,
			proc_read_vlan, NULL);
	create_proc_read_entry("port_status", 0444, adm5120sw_dir,
			proc_read_all_port_status, NULL);

	for (i = 0; i < NUM_IF5120_PORTS; i++) {
		char buf[16];

		sprintf(buf, "%d", i);
		adm5120sw_port_dir[i] = proc_mkdir(buf, adm5120sw_dir);
		if (adm5120sw_port_dir[i] == NULL)
			continue;

		for (j = 0; proc_port_entries[j].name; j++) {
			struct port_entry *port = &proc_port_entries[j];
			struct proc_dir_entry *node;

			node = create_proc_entry(port->name, port->mode,
					adm5120sw_port_dir[i]);
			if (node == NULL)
				continue;
			node->read_proc = port->read;
			node->write_proc = port->write;
			node->data = (void *)_PROC_ID(i, port->entry);
		}
	}
#endif

	return 0;
}



#ifdef CONFIG_PROC_FS
static int proc_read_vlan(char *page, char **start, off_t offset,
			  int len, int *eof, void *data)
{
	int i, j;
	int n = 0;
	unsigned ports;

	for (i = 0; i < MAX_VLAN_GROUP; i++) {
		ports = sw_context->vlanGrp[i];
		if (ports == 0)
			break;

		n += snprintf(page + n, PAGE_SIZE - n, "eth%d: ", i);
		for (j = 0; j < IF5120_PORT_NUM; j++) {
			if (ports & 0x1)
				n += snprintf(page + n, PAGE_SIZE - n, "%d ", j);
			ports >>= 1;
		}
		n += snprintf(page + n, PAGE_SIZE - n, "\n");
	}

	*start = page + offset;
	n -= offset;
	if (n < 0)
		n = 0;
	if (n <= len)
		*eof = 1;
	if (n > len)
		n = len;
		
	return n;
}


static int proc_read_all_port_status(char *page, char **start, off_t offset,
			  int len, int *eof, void *data)
{
	int i;
	int n = 0;

	for (i = 0; i < NUM_IF5120_PORTS; i++) {
		n += snprintf(page + n, PAGE_SIZE - n, "port%d: ", i);
		if (sw_context->port[i].status == PORT_ENABLED)
			n += snprintf(page + n, PAGE_SIZE - n, "enabled, ");
		else {
			n += snprintf(page + n, PAGE_SIZE - n, "disabled\n");
			continue;
		}
		n += snprintf(page + n, PAGE_SIZE - n, "vlan:%d, ",
				sw_context->port[i].ifUnit);
		n += snprintf(page + n, PAGE_SIZE - n, "link:%s, ",
				(sw_context->port[i].link) ? "up" : "down");
		n += snprintf(page + n, PAGE_SIZE - n, "speed:");
		switch (sw_context->port[i].speed) {
		case 0:
			n += snprintf(page + n, PAGE_SIZE - n, "10Mbps, ");
			n += snprintf(page + n, PAGE_SIZE - n,
				"flow-control:%s, ",
				(sw_context->port[i].flow_control) ?
					"on" : "off");
			break;
		case 1:
			n += snprintf(page + n, PAGE_SIZE - n, "100Mbps, ");
			n += snprintf(page + n, PAGE_SIZE - n,
				"flow-control:%s, ",
				(sw_context->port[i].flow_control)?
					"on" : "off");
			break;
		case 3:
			n += snprintf(page + n, PAGE_SIZE - n, "1000Mbps, ");
			n += snprintf(page + n, PAGE_SIZE - n,
				"TX-flow-control:%s, ",
				((sw_context->port[i].flow_control) & 2) ?
			    		"on" : "off");
			n += snprintf(page + n, PAGE_SIZE - n,
				"RX-flow-control:%s, ",
				(sw_context->port[i].flow_control) ?
					"on" : "off");
			break;
		}
		n += snprintf(page + n, PAGE_SIZE - n, "duplex:%s\n",
			(sw_context->port[i].duplex) ? "full" : "half");
	}
	*start = page + offset;
	n -= offset;
	if (n < 0)
		n = 0;
	if (n <= len)
		*eof = 1;
	if (n > len)
		n = len;
		
	return n;
}


static int
proc_read_port_status(char *page, char **start, off_t offset,
			  int len, int *eof, void *data)
{
	int portnum = _PROC_ID_PORT((int)data);
	int value;
	int n;

	switch (_PROC_ID_ENTRY((int)data)) {
	case _STATUS_ENABLE:
		value = sw_context->port[portnum].status;
		break;
	case _STATUS_LINK:
		value = sw_context->port[portnum].link;
		break;
	case _STATUS_SPEED:
		value = sw_context->port[portnum].speed;
		break;
	case _STATUS_DUPLEX:
		value = sw_context->port[portnum].duplex;
		break;
	case _STATUS_FC:
		value = sw_context->port[portnum].flow_control;
		break;
	case _STATUS_VLAN:
		value = sw_context->port[portnum].ifUnit;
		break;
	default:
		return -ENOENT;
	}

	n = snprintf(page, PAGE_SIZE, "%d", value);
	
	*start = page + offset;
	n -= offset;
	if (n < 0)
		n = 0;
	if (n <= len)
		*eof = 1;
	if (n > len)
		n = len;
		
	return n;
}


static int
proc_read_port_control(char *page, char **start, off_t offset,
			  int len, int *eof, void *data)
{
	int portnum = _PROC_ID_PORT((int)data);
	int entry = _PROC_ID_ENTRY((int)data);
	int value;
	int n;

	if (portnum != 5) {
		unsigned long control;
		control = ADM5120_SW_REG(PHY_cntl2_REG);
		switch (entry) {
		case _CONTROL_AN:
			value = (control >> portnum) & 1;
			break;
		case _CONTROL_SPEED:
			value = (control >> (5 + portnum)) & 1;
			break;
		case _CONTROL_DUPLEX:
			value = (control >> (10 + portnum)) & 1;
			break;
		case _CONTROL_FC:
			value = (control >> (15 + portnum)) & 1;
			break;
		case _CONTROL_MDIX:
			value = (control >> (25 + portnum)) & 1;
			break;
		default:
			return -ENOENT;
		}
	} else {
		/* GMII */
		unsigned long conf;
		conf = ADM5120_SW_REG(Port_conf2_REG);
		switch (entry) {
		case _CONTROL_AN:
			value = conf & 1;
			break;
		case _CONTROL_SPEED:
			value = (conf >> 1) & 3;
			break;
		case _CONTROL_DUPLEX:
			value = (conf >> 3) & 1;
			break;
		case _CONTROL_FC:
			value = (conf >> 4) & 3;
			break;
		case _CONTROL_MDIX:
			value = (conf >> 7) & 1;
			break;
		default:
			return -ENOENT;
		}
	}

	n = snprintf(page, PAGE_SIZE, "%d", value);

	*start = page + offset;
	n -= offset;
	if (n < 0)
		n = 0;
	if (n <= len)
		*eof = 1;
	if (n > len)
		n = len;
		
	return n;
}



static int
proc_write_port_control(struct file *file, const char *userbuf,
				 unsigned long count, void *data)
{
	unsigned long flags;
	int error;
	int portnum = _PROC_ID_PORT((int)data);
	int entry = _PROC_ID_ENTRY((int)data);
	char buf[16], *p;
	int value;
	int n;
	
	if (count == 0)
		return 0;

	n = count;
	if (n > sizeof(buf) - 1)
		n = sizeof(buf) - 1;

	error = copy_from_user(buf, userbuf, n);
	if (error)
		return error;

	buf[n] = 0;

	value = simple_strtoul(buf, &p, 10);
	if (*p != 0 && !isspace(*p))
		return -EINVAL;

	if (portnum < 5) {
		int shift;
		unsigned long mask, bit;

		if (value > 1)
			return -EINVAL;

		switch (entry) {
		case _CONTROL_AN:
			shift = 0;
			break;
		case _CONTROL_SPEED:
			shift = 5;
			break;
		case _CONTROL_DUPLEX:
			shift = 10;
			break;
		case _CONTROL_FC:
			shift = 15;
			break;
		case _CONTROL_MDIX:
			shift = 25;
			break;
		default:
			return -ENOENT;
		}

		mask = ~(1 << (shift + portnum));
		bit = value << (shift + portnum);

		spin_lock_irqsave(&sw_context->lock, flags);

		ADM5120_SW_REG(PHY_cntl2_REG) =
			(ADM5120_SW_REG(PHY_cntl2_REG) & mask) | bit;

		spin_unlock_irqrestore(&sw_context->lock, flags);
	} else {
		/* GMII */
		unsigned long mask, bit;

		switch (entry) {
		case _CONTROL_AN:
			mask = 1;
			bit = value;
			break;
		case _CONTROL_SPEED:
			mask = 3 << 1;
			bit = value << 1;
			break;
		case _CONTROL_DUPLEX:
			mask = 1 << 3;
			bit = value << 3;
			break;
		case _CONTROL_FC:
			mask = 3 << 4;
			bit = value << 4;
			break;
		case _CONTROL_MDIX:
			mask = 1 << 4;
			bit = value << 4;
			break;
		default:
			return -ENOENT;
		}

		spin_lock_irqsave(&sw_context->lock, flags);

		ADM5120_SW_REG(Port_conf2_REG) =
			(ADM5120_SW_REG(Port_conf2_REG) & mask) | bit;

		spin_unlock_irqrestore(&sw_context->lock, flags);
	}

	UpdateLinkStatus();

	return n;
}

#endif


/*
 * module cleanup
 */
static void __exit adm5120switch_cleanup(void)
{
	int i;

	for (i = 0; i < nvlangroups; i++) {
		kfree(adm5120sw_devs[i].priv);
		unregister_netdev(adm5120sw_devs + i);
	}
}


module_init(adm5120switch_init);
module_exit(adm5120switch_cleanup);


