/*****************************************************************************
;
;   (C) ADMtek Incorporated.  All Rights Reserved.
;
;------------------------------------------------------------------------------
;
;    Project : ADM5120
;    Creator : daniell@admtek.com.tw
;    File    : net/adm5120sw/swdrv.c
;	 Date    : 2003.5.14
;    Abstract: ADM5120 Switch driver
;
;Modification History:
;
;*****************************************************************************/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/delay.h>

#include <linux/sched.h>
#include <linux/kernel.h>		/* printk() */
#include <linux/slab.h>			/* kmalloc() */
#include <linux/errno.h>		/* error codes */
#include <linux/types.h>		/* size_t */
#include <linux/interrupt.h>	/* mark_bh */

#include <linux/in.h>
#include <linux/netdevice.h>	/* struct device, and other headers */
#include <linux/etherdevice.h>	/* eth_type_trans */
#include <linux/ip.h>			/* struct iphdr */
#include <linux/tcp.h>			/* struct tcphdr */
#include <linux/skbuff.h>
#include <linux/in6.h>

#include <asm/checksum.h>

#include "if5120.h"


/*
 * Driver resources constants
 */
#define NUM_TX_H_DESC		16		/* Number of the Transmitting descriptors of high priority */
#define NUM_TX_L_DESC		128		/* Number of the Transmitting descriptors of low priority */
#define NUM_RX_H_DESC		16		/* Number of the Receiving descriptors of high priority */
#define NUM_RX_L_DESC		128		/* Number of the Receiving descriptors of low priority */

PSW_CONTEXT_T sw_context;

extern struct net_device adm5120sw_devs[ADM5120SW_IFNUM_MAX];


/* Set Promisc mode
 * Daniel+ 2003.8.4
 */
void EnablePromisc(int unit, unsigned long rxmode)
{
	int port = sw_context->vlanGrp[unit] & 0x3f;

	if (rxmode == IF_PROMISC_MODE) {
		ADM5120_SW_REG(CPUp_conf_REG) &= ~(port << 9);
		ADM5120_SW_REG(Port_conf1_REG) |= (port) | (port << 6);
		ADM5120_SW_REG(Port_conf1_REG) &= ~(port << 20);
	} else {
		ADM5120_SW_REG(CPUp_conf_REG) |= (port << 9);
		ADM5120_SW_REG(Port_conf1_REG) &= ~(port) | ~(port << 6);
		ADM5120_SW_REG(Port_conf1_REG) |= (port << 20);
	}
	adm_sync();
}


/*
 * Enable VLAN Group
 */
void EnableVlanGroup(int unit, unsigned long vlanGrp)
{
	int i;
	
	vlanGrp &= SW_DISABLE_PORT_MASK;
	ADM5120_SW_REG(Port_conf0_REG) &= ~vlanGrp; 
	adm_sync();

	/* Mark the enabled ports */
	for (i = 0; i < NUM_IF5120_PORTS; i++)
	{
		if (vlanGrp & (0x01 << i)) 
		{
			sw_context->port[i].status = PORT_ENABLED;
			sw_context->port[i].ifUnit = unit;
		}
	}
}


/*
 * DisableVlanGroup
 */
void DisableVlanGroup(int unit, unsigned long vlanGrp)
{
	int i;
	unsigned long reg;
	
	vlanGrp &= SW_DISABLE_PORT_MASK;
	
	reg = ADM5120_SW_REG(Port_conf0_REG) | vlanGrp; 
	ADM5120_SW_REG(Port_conf0_REG) = reg;
	adm_sync();

	/* Mark the disabled ports */
	for (i = 0; i < NUM_IF5120_PORTS; i++)
		if (vlanGrp & (0x01<<i)) 
			sw_context->port[i].status = PORT_DISABLED;
}


/*
 * Program VLAN mac
 */
int ProgramVlanMac(int vlan, unsigned char *Mac, int clear)
{
	unsigned long Reg0, Reg1;
	
	if (vlan < 0 || vlan >= MAX_VLAN_GROUP) 
		return -1;
	
	Reg0 = (((unsigned long)Mac[1] << 24) + ((unsigned long)Mac[0] << 16)) |
		 (vlan << SW_MAC_VLANID_SHIFT) |
		SW_MAC_WRITE | SW_MAC_VLANID_EN;

	if (!clear)
		Reg0 |= SW_MAC_AGE_VALID;
	
	Reg1 = ((unsigned long)Mac[5] << 24) + ((unsigned long)Mac[4] << 16) + 
		((unsigned long)Mac[3] << 8) + Mac[2];

	ADM5120_SW_REG(MAC_wt1_REG) = Reg1;
	ADM5120_SW_REG(MAC_wt0_REG) = Reg0;

	adm_sync();

	while (!(ADM5120_SW_REG(MAC_wt0_REG) & SW_MAC_WRITE_DONE));

	return 0;
}


/*
 * Setup VLAN
 */
int SetupVLAN(int unit, unsigned long portmask)
{
	unsigned long reg, shiftcnt;
	
	if (unit < 0 || unit > 6) 
		return -1;
	
	if (unit <= 3)
	{
		shiftcnt = 8 * unit;
		reg = ADM5120_SW_REG(VLAN_G1_REG) & ~(VLAN_PORT_MASK << shiftcnt);
		reg |= (portmask & VLAN_PORT_MASK) << shiftcnt;
		ADM5120_SW_REG(VLAN_G1_REG) = reg;
	}
	else
	{
		shiftcnt = 8 * (unit - 4);
		reg = ADM5120_SW_REG(VLAN_G2_REG) & ~(VLAN_PORT_MASK << shiftcnt);
		reg |= (portmask & VLAN_PORT_MASK) << shiftcnt;
		ADM5120_SW_REG(VLAN_G2_REG) = reg;
	}
	adm_sync();

	return 0;
}


/*
 * ProcessRxInt
 */
void ProcessRxInt(PRX_ENG_T rxEng)
{
	struct net_device *rdev;
	PSW_PRIV_T priv = 0;
	PRXDESC_T rxDesc = 0;
	PRX_DRV_DESC_T drvDesc = 0;
	int unit;
	int srcPort;
	int idx;
	int len;

	idx = rxEng->idx;
	rxDesc = &rxEng->hwDesc[idx];
	while (!(rxDesc->buf1cntl & OWN_BIT))
	{
		drvDesc = &rxEng->drvDesc[idx];
		if (drvDesc->skb == 0)
			goto get_desc;

		srcPort = (rxDesc->status & RX_SRC_PORT_MASK)
		    >> RX_SRC_PORT_SHIFT;
		unit = sw_context->port[srcPort].ifUnit;
		rdev = &adm5120sw_devs[unit];
		priv = (PSW_PRIV_T)rdev->priv;

		len = ((rxDesc->status & PKT_LEN_MASK) >> PKT_LEN_SHIFT)
		    - ETH_CRC_LEN;
		if (len <= 0 || (rxDesc->status & RX_PKT_IPSUM_ERR)) {
			priv->stats.rx_errors++;
			priv->stats.rx_length_errors++;
			goto next;
		}
		
		skb_put(drvDesc->skb, len);
			
		/* Write metadata, and then pass to the receive level */
		drvDesc->skb->dev = rdev;
		drvDesc->skb->protocol = eth_type_trans(drvDesc->skb, rdev);
		drvDesc->skb->ip_summed = CHECKSUM_UNNECESSARY; /* don't check it */

#ifdef GPIO_MEASURE
		GPIO_TOGGLE(1);
#endif
		if (netif_rx(drvDesc->skb) == NET_RX_DROP)
		{
			priv->stats.rx_dropped++;
		}
		else
		{
			rdev->last_rx = jiffies;
			priv->stats.rx_packets++;
			priv->stats.rx_bytes += drvDesc->skb->len;
		}

#ifdef GPIO_MEASURE
		GPIO_TOGGLE(1);
#endif

get_desc:

#ifdef GPIO_MEASURE
		GPIO_TOGGLE(1);
#endif
		drvDesc->skb = __dev_alloc_skb(DEF_RX_BUF_SIZE+16,
				GFP_ATOMIC | GFP_DMA);

#ifdef GPIO_MEASURE
		GPIO_TOGGLE(1);
#endif

		if (drvDesc->skb) 
		{
			skb_reserve(drvDesc->skb, 2); /* align IP on 16B boundary */
next:
			rxDesc->buf2cntl = rxDesc->status = 0;
			rxDesc->buf1len = DEF_RX_BUF_SIZE;
			rxDesc->buf1cntl = (rxDesc->buf1cntl & END_OF_RING) | OWN_BIT
							| (((unsigned long)drvDesc->skb->data)& BUF_ADDR_MASK);
		}
		else
		{
			printk("Init rx skb : low on mem\n");
		}
			
		if (++idx == rxEng->numDesc)
		{
			idx = 0;
		}

		rxDesc = &rxEng->hwDesc[idx];
	}

	rxEng->idx = idx;

    return;
}


/*
 * ProcessTxInt
 */
void ProcessTxInt(PTX_ENG_T txEng)
{
	PTX_DRV_DESC_T drvDesc;
	int idx;

	idx = txEng->idxTail;

	while (!(txEng->hwDesc[idx].buf1cntl & OWN_BIT) && (idx != txEng->idxHead))
	{
		drvDesc = &txEng->drvDesc[idx];
#ifdef GPIO_MEASURE
	GPIO_TOGGLE(3);
#endif
		dev_kfree_skb_irq(drvDesc->skb);

#ifdef GPIO_MEASURE
	GPIO_TOGGLE(3);
#endif	
		drvDesc->skb = 0;

		if (++idx == txEng->numDesc)
			idx = 0;
	}

	txEng->idxTail = idx;
}


/*
 * You must lock sw_context->lock while calling UpdateLinkStatus().
 */
void UpdateLinkStatus(void)
{
	int i;
	unsigned phy_st = ADM5120_SW_REG(PHY_st_REG);

	for (i = 0; i < NUM_IF5120_PORTS; i++) {
		sw_context->port[i].link = (phy_st >> i) & 1;
		sw_context->port[i].duplex = (phy_st >> (16 + i)) & 1;
		if (i < 5) {
			sw_context->port[i].speed = (phy_st >> (8 + i)) & 1;
			sw_context->port[i].flow_control =
				(phy_st >> (24 + i)) & 1;
		} else {
			/* GMII port */
			sw_context->port[i].speed = (phy_st >> 13) & 3;
			sw_context->port[i].flow_control = (phy_st >> 29) & 3;
		}
#if 0
		printk("adm5120sw port%d: ", i);
		printk("link=%d speed=%d duplex=%d flow_control=%d\n",
				sw_context->port[i].link,
				sw_context->port[i].speed,
				sw_context->port[i].duplex,
				sw_context->port[i].flow_control);
#endif
	}

	/* 
	 * determine vlan port link status as forrowings
	 *
	 * link up:	one or more ports in the group are up.
	 * link down:	all vlan ports are down.
	 *
	 * 100baseTX:	all linked ports are 100baseTX.
	 * 10baseT:	one or more ports are 10baseT.
	 * unknown:	only GMII port is linked and the speed is GIGA.
	 */
	for (i = 0; i < sw_context->nvlanGrps; i++) {
		struct net_device *dev;
		unsigned ports = sw_context->vlanGrp[i] & 0x3f;
		unsigned link;
		int speed;

		dev = &adm5120sw_devs[i];
		link = phy_st & ports;
		if (link != 0) {
			if (!netif_carrier_ok(dev)) {
				printk(KERN_INFO
					"adm5120_eth vlan %d link up\n", i);
				netif_carrier_on(dev);
			}
			if (link == 0x20) {
				/* only GMII port is linked */
				speed = sw_context->port[5].speed;
			} else {
				unsigned ports_10M;

				ports_10M = (~(phy_st >> 8)) & 0x1f;
				if (((phy_st >> 13) & 0x3) == 0x0)
					ports_10M |= 0x20;
				if ((ports_10M & link) == link) {
					/* all linked ports are 10baseT */
					speed = 0;
				} else
					speed = 1;
			}
			switch (speed) {
			case 0:
				dev->if_port = IF_PORT_10BASET;
				break;
			case 1:
				dev->if_port = IF_PORT_100BASETX;
				break;
			case 2:
				dev->if_port = IF_PORT_UNKNOWN;
			}
		} else {
			if (netif_carrier_ok(dev)) {
				printk(KERN_INFO
					"adm5120_eth vlan %d link down\n", i);
				netif_carrier_off(dev);
				dev->if_port = 0;
			}
		}
	}
}


/*
 * swdrv_ProcessInt
 */
void swdrv_ProcessInt(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned long intReg;
	
	spin_lock(&sw_context->lock);

#ifdef GPIO_MEASURE
	GPIO_TOGGLE(0);
#endif

	/* disable switch interrupt */
	ADM5120_SW_REG(SW_Int_mask_REG) |= sw_context->intMask;
	adm_sync();

	/* recording the current interrupts in the context. */	
	intReg = ADM5120_SW_REG(SW_Int_st_REG);
	
	/* acknowledge  all interrupts */
	ADM5120_SW_REG(SW_Int_st_REG ) = intReg;
	adm_sync();
	
	/* receive one high priority packet to cpu */
	if (intReg & RX_H_INT)	
		ProcessRxInt(&sw_context->rxH);
		
	/* receive one nromal priority packet to cpu */
	if (intReg & RX_L_INT)	
		ProcessRxInt(&sw_context->rxL);

	/* transmit one high priority packet to cpu */
	if (intReg & TX_H_INT)	
		ProcessTxInt(&sw_context->txH);

	/* transmit one nromal priority packet to cpu */
	if (intReg & TX_L_INT)	
		ProcessTxInt(&sw_context->txL);

	/* check link status */
	if (intReg & LINK_INT)
		UpdateLinkStatus();

	ADM5120_SW_REG(SW_Int_mask_REG) &= ~sw_context->intMask;
	adm_sync();

#ifdef GPIO_MEASURE
	GPIO_TOGGLE(0);
#endif

	spin_unlock(&sw_context->lock);
}


/*
 * swdrv_HwTx
 */
int swdrv_HwTx(struct sk_buff *skb, PSW_PRIV_T priv)
{
	PTX_ENG_T txEng;
	PTXDESC_T hdesc;
	PTX_DRV_DESC_T drvDesc;
	unsigned long csum = 0;
	int len;

	/* get tx engine */
	if (priv->priority == IF_TX_PRIORITY_H)
		txEng = &sw_context->txH;
	else
		txEng = &sw_context->txL;
	
	/* get check sum flag */
	if (priv->csum_flags)
		csum = TX_ADD_IPSUM;

	/* get hardware descriptor */
	hdesc = &txEng->hwDesc[txEng->idxHead];
	if (hdesc->buf1cntl & OWN_BIT)
	{
		dev_kfree_skb(skb);
		priv->stats.tx_dropped++;
		return 0;
	}
	if ((PHYSADDR(skb->data) & ~BUF_ADDR_MASK) != 0) {
		struct sk_buff *newskb;
#if 0
		printk("the skb is unreachable (%p), move to lower memory\n", skb->data);
#endif
		newskb = skb_copy(skb, GFP_DMA | GFP_ATOMIC);
		if (newskb == NULL) {
			printk("failed on skb_copy()\n");
			dev_kfree_skb_irq(skb);
			priv->stats.tx_dropped++;
			return 0;
		}
		dev_kfree_skb_irq(skb);
		skb = newskb;
	}

	drvDesc = &txEng->drvDesc[txEng->idxHead];
	drvDesc->skb = skb;
	
	hdesc->buf1cntl = (hdesc->buf1cntl & END_OF_RING) | 
		 	(((unsigned long)skb->data & BUF_ADDR_MASK) | OWN_BIT);
		
	len = skb->len < ETH_ZLEN ? ETH_ZLEN : skb->len;
	hdesc->pktcntl = (len << PKT_LEN_SHIFT) | priv->vlanId | csum;
	hdesc->buf1len = len;

	priv->stats.tx_packets++;
	priv->stats.tx_bytes += len;

	if (++txEng->idxHead >= txEng->numDesc)
		txEng->idxHead = 0;

	ADM5120_SW_REG(Send_trig_REG) = txEng->txTrig;
	adm_sync();

	return 0;
}



/*
 * switch engine init
 */
void engine_init(void)
{
	int i;

	/* disable cpu port, CRC padding from cpu and 
	   no send unknown packet from port0 to port5 to cpu */
	ADM5120_SW_REG(CPUp_conf_REG) = DEF_CPUPORT_CFG	;

	/* Disable all port, enable BP & MC */
	ADM5120_SW_REG(Port_conf0_REG) = DEF_PORTS_CFG;

	adm_sync();

   	/* Wait untile switch enter idle state */
  	for (i=0; i < 1000; i++)
		udelay(100);
  	
	/* Put Phys to normal mode */
	ADM5120_SW_REG(PHY_cntl2_REG) |= SW_PHY_AN_MASK | SW_PHY_NORMAL_MASK;
	ADM5120_SW_REG(PHY_cntl3_REG) = 0x400;	/* XXX */

	/* Disable Switch Interrupts */
	ADM5120_SW_REG(SW_Int_mask_REG) = SWITCH_INT_MASK;

	/* Clear the Interrupt status */
	ADM5120_SW_REG(SW_Int_st_REG) = SWITCH_INT_MASK;

	/* Initialize the adm5120 Desc */
	ADM5120_SW_REG(Send_HBaddr_REG) = (unsigned long)sw_context->txH.hwDesc;
	ADM5120_SW_REG(Send_LBaddr_REG) = (unsigned long)sw_context->txL.hwDesc;
	ADM5120_SW_REG(Recv_HBaddr_REG) = (unsigned long)sw_context->rxH.hwDesc;
	ADM5120_SW_REG(Recv_LBaddr_REG) = (unsigned long)sw_context->rxL.hwDesc;

	/* Clear all vlan setting */
	ADM5120_SW_REG(VLAN_G1_REG) = 0;
	ADM5120_SW_REG(VLAN_G2_REG) = 0;

	sw_context->intMask = RX_H_INT | RX_L_INT | TX_H_INT | TX_L_INT | LINK_INT;

	ADM5120_SW_REG(CPUp_conf_REG) &= ~SW_CPU_PORT_DISABLE;

	adm_sync();
}


/* 
 * InitTxDesc
 */															
static void InitTxDesc(PTX_ENG_T pTxEng)
{
	int num = pTxEng->numDesc;

	pTxEng->hwDesc[--num].buf1cntl |= END_OF_RING;
	pTxEng->idxHead = pTxEng->idxTail = 0;
}


/* 
 * InitRxDesc
 */									
static void InitRxDesc(PRX_ENG_T pRxEng)
{
	PRX_DRV_DESC_T drvDesc = pRxEng->drvDesc;
	int i;

	for (i = 0; i < pRxEng->numDesc; i++, drvDesc++)
	{
		drvDesc->skb = __dev_alloc_skb(DEF_RX_BUF_SIZE+16, GFP_DMA);
		if (!drvDesc->skb) 
		{
			printk("Init rx skb : low on mem\n");
			return;
		}
		skb_reserve(drvDesc->skb, 2); /* align IP on 16B boundary */
	}

	drvDesc = pRxEng->drvDesc;

	for (i = 0; i < pRxEng->numDesc; i++)
	{
		pRxEng->hwDesc[i].buf2cntl = pRxEng->hwDesc[i].status = 0;
		pRxEng->hwDesc[i].buf1len = DEF_RX_BUF_SIZE;
		pRxEng->hwDesc[i].buf1cntl = 
			((unsigned long)pRxEng->drvDesc[i].skb->data & 
			BUF_ADDR_MASK) | OWN_BIT;
	}

	pRxEng->hwDesc[--i].buf1cntl |= END_OF_RING;
	pRxEng->idx = 0;
}


/*
 * Switch driver init
 */
int swdrv_init(int nvlangroups, unsigned char *vlangroups)
{
	int i;

	/* Allocate the switch driver context */
	if ((sw_context = (PSW_CONTEXT_T)kmalloc(sizeof(SW_CONTEXT_T), GFP_KERNEL)) == NULL) 
		return (-1);

	memset((char *)sw_context, 0, sizeof(SW_CONTEXT_T));
	for (i = 0; i < nvlangroups; i++)
		sw_context->vlanGrp[i] = vlangroups[i];
	sw_context->nvlanGrps = i;

	/* Allocate the Tx/Rx hardware descriptor pool */
	i = HWDESC_ALIGN + sizeof(RXDESC_T) * (NUM_RX_H_DESC + NUM_RX_L_DESC)
		+ sizeof(TXDESC_T) * (NUM_TX_H_DESC + NUM_TX_L_DESC);

	if ((sw_context->hwDescPool = (char *)kmalloc(i, GFP_KERNEL|GFP_DMA)) == NULL) 
		goto ErrRes;

	memset(sw_context->hwDescPool, 0, i); 

	/* Allocate the tx driver descriptor */
	i = sizeof(TX_DRV_DESC_T) * (NUM_TX_H_DESC + NUM_TX_L_DESC);

	if ((sw_context->txDrvDescPool = (char *)kmalloc(i, GFP_KERNEL|GFP_DMA)) == NULL) 
		goto ErrRes;

	memset(sw_context->txDrvDescPool, 0, i); 

	/* Allocate the rx driver descriptor */
	i = sizeof(RX_DRV_DESC_T) * (NUM_RX_H_DESC + NUM_RX_L_DESC);

	if ((sw_context->rxDrvDescPool = (char *)kmalloc(i, GFP_KERNEL|GFP_DMA)) == NULL) 
		goto ErrRes;

	memset(sw_context->rxDrvDescPool, 0, i); 

	/* 
	 *!! Note: The Hardware Tx/Rx descriptors should be allocated at non-cached memory and 
	 * 			aligned on 16 bytes boundry!!!!!!
	 */

	/* assign hardware descriptor to txH pool */
	if (((unsigned long)sw_context->hwDescPool) & (HWDESC_ALIGN - 1))
	{
		sw_context->txH.hwDesc = (PTXDESC_T)KSEG1ADDR(
		    (unsigned long)(sw_context->hwDescPool + HWDESC_ALIGN - 1) &
		     ~(HWDESC_ALIGN - 1));
	}
	else
		sw_context->txH.hwDesc = (PTXDESC_T)KSEG1ADDR(
		    sw_context->hwDescPool);
	
	sw_context->txH.numDesc = NUM_TX_H_DESC;
	sw_context->txH.txTrig = SEND_TRIG_HIGH;
	sw_context->txH.drvDesc = (PTX_DRV_DESC_T)sw_context->txDrvDescPool;
	InitTxDesc(&sw_context->txH);

	/* assign hardware descriptor to txL pool */
	sw_context->txL.hwDesc = &sw_context->txH.hwDesc[NUM_TX_H_DESC];
	sw_context->txL.numDesc = NUM_TX_L_DESC;
	sw_context->txL.txTrig = SEND_TRIG_LOW;
	sw_context->txL.drvDesc = (PTX_DRV_DESC_T)(sw_context->txDrvDescPool + 
		                        (sizeof(TX_DRV_DESC_T) * NUM_TX_H_DESC));
	InitTxDesc(&sw_context->txL);

	/* assign hardware descriptor to rxH pool */
	sw_context->rxH.hwDesc = (PRXDESC_T)&sw_context->txL.hwDesc[NUM_TX_L_DESC];
	sw_context->rxH.numDesc = NUM_RX_H_DESC;
	sw_context->rxH.drvDesc = (PRX_DRV_DESC_T)sw_context->rxDrvDescPool;
	InitRxDesc(&sw_context->rxH);

	/* assign hardware descriptor to rxL pool */
	sw_context->rxL.hwDesc = &sw_context->rxH.hwDesc[NUM_RX_H_DESC];
	sw_context->rxL.numDesc = NUM_RX_L_DESC;
	sw_context->rxL.drvDesc = (PRX_DRV_DESC_T)(sw_context->rxDrvDescPool + 
				(sizeof(RX_DRV_DESC_T) * NUM_RX_H_DESC));
	InitRxDesc(&sw_context->rxL);

	spin_lock_init(&sw_context->lock);

	/* init switch engine */
	engine_init();

#ifdef GPIO_MEASURE
	GPIO_MEASURE_INIT();

	GPIO_SET_LOW(0);
	GPIO_SET_LOW(1);
	GPIO_SET_LOW(2);
	GPIO_SET_LOW(3);
#endif

	return 0;

ErrRes:
	/* Free all resources */ 
	if (sw_context->hwDescPool != NULL)
		kfree(sw_context->hwDescPool);

	if (sw_context->txDrvDescPool != NULL)
		kfree(sw_context->txDrvDescPool);

	if (sw_context->rxDrvDescPool != NULL)
		kfree(sw_context->rxDrvDescPool);

	kfree(sw_context);

	return (-1);
}


